structure DeflateBits :> DeflateBits =
struct

	(* bool <-> int *)
	
	fun bit_to_int true = 1
	  | bit_to_int false = 0

	fun bit_to_char true = #"1"
	  | bit_to_char false = #"0"

	fun least_bit c = if c mod 2 = 1 then true else false

	(* int <-> bool list *)
	
	fun explode_fixed_lsb_first 0 _ = []
	  | explode_fixed_lsb_first len code = 
	  		least_bit code :: explode_fixed_lsb_first (len - 1) (code div 2)

	fun explode_fixed_msb_first len code = rev (explode_fixed_lsb_first len code)
	
	fun explode_byte b = explode_fixed_lsb_first 8 b

	fun to_int_lsb_first bits =
		let
			fun f (b, result) = (2 * result) + (bit_to_int b)
		in
			foldr f 0 bits
		end

	fun to_int_msb_first bits = to_int_lsb_first (rev bits)
	
	fun to_string bits =
		implode (map bit_to_char bits)

	(* power of 2 *)
	
	fun pow2 0 = 1 
	  | pow2 n =
	  		if (n mod 2) = 1 then 2 * pow2 (n - 1)
			else let val p = pow2 (n div 2) in p * p end

	(* bit-iterable byte stream *)

	type byteio = bool list * int list * int list
	(*
		#1: bits that have been extracted from input bytes
		#2: unprocessed input bytes
		#3: output bytes
	*)
	
	exception Unexpected_end_of_stream
	exception Unaligned_byte_operation

	fun io_to_string (bits, bytes, result) = 
		"(" ^ Int.toString (length bits) ^ " bits, " 
		^ Int.toString (length bytes) ^ " in, "
		^ Int.toString (length result) ^ " out)"
		
	(* get one more byte into bits *)
	
	fun from_input bytes = ([], bytes, [])
	
	fun forth (bits, b1 :: bs, result) = (bits @ (explode_byte b1), bs, result)
	  | forth _ = raise Unexpected_end_of_stream

	(* append byte(s) to output *)

	fun append_bytes (bits, bs, result) rs = (bits, bs, rev result @ rs) 

	fun append_byte (bits, bs, result) r = (bits, bs, r :: result)
	 		
	(* take/drop bytes from input *)

	fun take_bytes (b::bs, _, _) _ = raise Unaligned_byte_operation
	  | take_bytes ([], bytes, _) nbytes = List.take (bytes, nbytes)
	  
	fun drop_bytes (b::bs, _, _) _ = raise Unaligned_byte_operation
	  | drop_bytes ([], bytes, result) nbytes = ([], List.drop (bytes, nbytes), result)
	
	fun take_zero_string (b::bs, _, _) = raise Unaligned_byte_operation
	  | take_zero_string ([], bytes, _) =
	  	let
			fun impl (0::_) = []
			  | impl (b::bs) = b :: impl bs
			  | impl [] = impl []
		in
			implode (map chr (impl bytes))
		end
	  
	fun drop_zero_string (b::bs, _, _) = raise Unaligned_byte_operation
	  | drop_zero_string ([], 0::bs, result) = ([], bs, result)
	  | drop_zero_string ([], _::bs, result) = drop_zero_string ([],bs,result)
	  | drop_zero_string ([], [], result) = ([],[],result)
		
	(* output *)
	
	fun append_from_output (bits, bytes, result) nbytes distance =
		let
			val taken = List.take (result, distance)
		in
			if nbytes > distance then
				append_from_output
					(bits, bytes, taken @ result)
					(nbytes - distance) distance
			else
				(bits, bytes, List.drop (taken, distance - nbytes) @ result)
		end
				
	fun output (_, _, result) = rev result
	
	(* take/drop bits from input *)
	
	fun take_bits (bits, bs, result) n =
		if length bits >= n then
			List.take (bits, n)
		else 
			take_bits (forth (bits, bs, result)) n

	fun drop_bits io 0 = io
	  | drop_bits ([], b::bs, result) n =
		if n >= 8 then
			drop_bits ([], bs, result) (n - 8)
		else
			drop_bits (explode_byte b, bs, result) n
	  | drop_bits (bits, bytes, result) n =
		let 
			val n_bits = length bits
		in
			if n_bits >= n then 
				(List.drop (bits, n), bytes, result)
			else 
				drop_bits ([], bytes, result) (n - n_bits)
		end

	fun drop_to_byte (_, bs, result) = ([], bs, result)
	
	(* get fixed size value *)
	fun get_fixed 0 iostate = (0, iostate)
  	  | get_fixed nbits iostate = (to_int_lsb_first (take_bits iostate nbits), drop_bits iostate nbits)
	
	(* self tests *)
	val tests = [
		least_bit 0 = false,
		least_bit 1 = true,
		least_bit 16 = false,
		least_bit 7 = true,
		bit_to_int true = 1,
		bit_to_int false = 0,
		explode_fixed_msb_first 6 15 = [false,false,true,true,true,true],
		to_int_lsb_first [true,true,false] = 3,
		to_string [false,true,true] = "011",
		to_string [] = "",
		forth ([true],[2,20,21],[4,5]) = ([true,false,true,false,false,false,false,false,false],[20,21],[4,5]),
		take_bits ([true,true,false],[2,4,8],[]) 2 = [true,true],
		take_bits ([false,true],[17,1,2],[]) 6 = [false,true,true,false,false,false],
		drop_bits ([true,true,false],[2,4,8],[2,1]) 2 = ([false],[2,4,8],[2,1]),
		drop_bits ([false,true],[17,1,2],[]) 6 = ([true,false,false,false],[1,2],[]),
		get_fixed 3 ([false,true,true,true],[1,2,3],[4,5]) = (6, ([true],[1,2,3],[4,5])),
		get_fixed 6 ([false,true],[33,1,2,3],[4,5]) = (4+2,([false,true,false,false],[1,2,3],[4,5]))
		]
end
