structure DeflateHuffmanBuild :> DeflateHuffmanBuild =
struct

	type bitcount = int * int * int 
	(* count codes 
		#1 length
		#2 count of code for this length
		#3 first code for this length 
	*)

	exception Unexpected_bit_length

	(* 3.2.2 #1 count the number of times each len is used *)

	fun count_bitlens bitlens = 
		let
			fun notzero i = i <> 0
			fun get_len (l,_,_) = l
			fun new_bitcount l = (l, 1, 0)
			fun increment_bitcount (len, count, startcode) = (len, count + 1, startcode)

			fun f (bitlen, []) = [new_bitcount bitlen]
			  | f (bitlen, r :: rs) =  
				if bitlen = get_len r then 
					increment_bitcount r :: rs
				else
					new_bitcount bitlen :: r ::rs
		in 
			foldr f [] (DeflateList.sort (op<=) (List.filter notzero bitlens))
		end

	(* 3.2.2 #2 find start code for each bit length *)

	fun build_startcodes bitlens =
		let
			fun startcode (_,_,code) = code
			fun has_hole (l1, _, _) (l2, _, _) = l1 + 1 < l2
			fun empty_next (l, _, _) = (l+1, 0, 0)
			
			fun count (_,c,_) = c
			fun filler (bc, []) = [bc]
			  | filler (newb, b::bs) =
				if has_hole b newb then
					filler (newb, empty_next b :: b :: bs)
				else
					newb :: b :: bs
			fun setter ((l,c,_), []) = [(l, c, 0)]
			  | setter ((l,c,_), b::bs) = (l, c, ((startcode b) + (count b)) * 2) :: b :: bs
		in
			rev (foldr setter [] (foldl filler [] (count_bitlens bitlens)))
		end

	(* 3.2.2 #3 build huffman tree *)

	fun from_lengths bitlens =
		let 
			val codes = (DeflateList.enum 0 (length bitlens - 1))
			fun code_for_length somecodes len = 
				case List.find (fn (l,_,_) => len = l) somecodes 
				of SOME (_, _, bc) => bc 
				| NONE => raise Unexpected_bit_length

			fun builder ((0, _), st) = st
			  | builder ((len,code), (startcodes,tree)) =
				let
					val huffman_position = code_for_length startcodes len
					val huffman_bits = DeflateBits.explode_fixed_msb_first len huffman_position
				in
					((len, 0, huffman_position + 1)::startcodes, 
						DeflateHuffman.insert tree huffman_bits code)
				end

			val (_, result) = foldl builder (build_startcodes bitlens, DeflateHuffman.none) (ListPair.zip (bitlens, codes))
		in
			result
		end
		
	(* self test *)
	
	val tests = [
		count_bitlens [1,1,1,2,2,3,3] = [(1,3,0),(2,2,0),(3,2,0)],
		count_bitlens [10,10,30,20,10,20,20] = [(10,3,0),(20,3,0),(30,1,0)],
		build_startcodes [3,3,3,3,3,2,4,4] = [(2,1,0),(3,5,2),(4,2,14)],
		(* 3.2.2 example *)
		DeflateHuffman.collapse_string (from_lengths [3,3,3,3,3,2,4,4]) =
				[("010",0),("011",1),("100",2),("101",3),
					("110",4),("00",5),("1110",6),("1111",7)],
		DeflateHuffman.collapse_string (from_lengths [3,3,0,0,3,3,3,2,0,4,4]) =
				[("010",0),("011",1), (* 0, 0 *)("100",4),("101",5),
					("110",6),("00",7), (* 0 *)("1110",9),("1111",10)]
		]
		
end
